# This file was auto-generated by Fern from our API Definition.

import typing
from json.decoder import JSONDecodeError

from ..core.api_error import ApiError
from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.http_response import AsyncHttpResponse, HttpResponse
from ..core.jsonable_encoder import jsonable_encoder
from ..core.pydantic_utilities import parse_obj_as
from ..core.request_options import RequestOptions
from ..core.serialization import convert_and_respect_annotation_metadata
from ..errors.bad_request_error import BadRequestError
from ..errors.conflict_error import ConflictError
from ..errors.gone_error import GoneError
from ..errors.not_found_error import NotFoundError
from ..types.error_message import ErrorMessage
from ..types.json_node import JsonNode
from ..types.local_runner import LocalRunner
from ..types.local_runner_connect_response import LocalRunnerConnectResponse
from ..types.local_runner_heartbeat_response import LocalRunnerHeartbeatResponse
from ..types.local_runner_job import LocalRunnerJob
from ..types.local_runner_job_metadata import LocalRunnerJobMetadata
from ..types.local_runner_job_page import LocalRunnerJobPage
from ..types.local_runner_log_entry import LocalRunnerLogEntry
from ..types.local_runner_page import LocalRunnerPage
from ..types.local_runner_pair_response import LocalRunnerPairResponse
from .types.local_runner_job_result_request_status import LocalRunnerJobResultRequestStatus

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class RawRunnersClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._client_wrapper = client_wrapper

    def get_job_logs(
        self,
        job_id: str,
        *,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[typing.List[LocalRunnerLogEntry]]:
        """
        Get log entries for a local runner job

        Parameters
        ----------
        job_id : str

        offset : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[typing.List[LocalRunnerLogEntry]]
            Log entries
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/jobs/{jsonable_encoder(job_id)}/logs",
            method="GET",
            params={
                "offset": offset,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    typing.List[LocalRunnerLogEntry],
                    parse_obj_as(
                        type_=typing.List[LocalRunnerLogEntry],  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def append_job_logs(
        self,
        job_id: str,
        *,
        request: typing.Sequence[LocalRunnerLogEntry],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Append log entries for a running local runner job

        Parameters
        ----------
        job_id : str

        request : typing.Sequence[LocalRunnerLogEntry]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/jobs/{jsonable_encoder(job_id)}/logs",
            method="POST",
            json=convert_and_respect_annotation_metadata(
                object_=request, annotation=typing.Sequence[LocalRunnerLogEntry], direction="write"
            ),
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def cancel_job(self, job_id: str, *, request_options: typing.Optional[RequestOptions] = None) -> HttpResponse[None]:
        """
        Cancel a pending or running local runner job

        Parameters
        ----------
        job_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/jobs/{jsonable_encoder(job_id)}/cancel",
            method="POST",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def connect_runner(
        self, *, pairing_code: str, runner_name: str, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[LocalRunnerConnectResponse]:
        """
        Exchange a pairing code or API key for local runner credentials

        Parameters
        ----------
        pairing_code : str

        runner_name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[LocalRunnerConnectResponse]
            Runner connected
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/local-runners/connections",
            method="POST",
            json={
                "pairing_code": pairing_code,
                "runner_name": runner_name,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerConnectResponse,
                    parse_obj_as(
                        type_=LocalRunnerConnectResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def create_job(
        self,
        *,
        agent_name: str,
        project_id: str,
        inputs: typing.Optional[JsonNode] = OMIT,
        mask_id: typing.Optional[str] = OMIT,
        metadata: typing.Optional[LocalRunnerJobMetadata] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Create a local runner job and enqueue it for execution

        Parameters
        ----------
        agent_name : str

        project_id : str

        inputs : typing.Optional[JsonNode]

        mask_id : typing.Optional[str]

        metadata : typing.Optional[LocalRunnerJobMetadata]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/local-runners/jobs",
            method="POST",
            json={
                "agent_name": agent_name,
                "inputs": inputs,
                "project_id": project_id,
                "mask_id": mask_id,
                "metadata": convert_and_respect_annotation_metadata(
                    object_=metadata, annotation=LocalRunnerJobMetadata, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def generate_pairing_code(
        self, *, project_id: str, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[LocalRunnerPairResponse]:
        """
        Generate a pairing code for a local runner in the current workspace

        Parameters
        ----------
        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[LocalRunnerPairResponse]
            Pairing code generated
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/local-runners/pairs",
            method="POST",
            json={
                "project_id": project_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerPairResponse,
                    parse_obj_as(
                        type_=LocalRunnerPairResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_job(
        self, job_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[LocalRunnerJob]:
        """
        Get a single local runner job's status and results

        Parameters
        ----------
        job_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[LocalRunnerJob]
            Job details
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/jobs/{jsonable_encoder(job_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerJob,
                    parse_obj_as(
                        type_=LocalRunnerJob,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def get_runner(
        self, runner_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[LocalRunner]:
        """
        Get a single local runner with its registered agents

        Parameters
        ----------
        runner_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[LocalRunner]
            Runner details
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/{jsonable_encoder(runner_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunner,
                    parse_obj_as(
                        type_=LocalRunner,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def heartbeat(
        self, runner_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[LocalRunnerHeartbeatResponse]:
        """
        Refresh local runner heartbeat

        Parameters
        ----------
        runner_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[LocalRunnerHeartbeatResponse]
            Heartbeat response
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/{jsonable_encoder(runner_id)}/heartbeats",
            method="POST",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerHeartbeatResponse,
                    parse_obj_as(
                        type_=LocalRunnerHeartbeatResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 410:
                raise GoneError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def list_jobs(
        self,
        runner_id: str,
        *,
        project_id: typing.Optional[str] = None,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[LocalRunnerJobPage]:
        """
        List jobs for a local runner

        Parameters
        ----------
        runner_id : str

        project_id : typing.Optional[str]

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[LocalRunnerJobPage]
            Jobs list
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/{jsonable_encoder(runner_id)}/jobs",
            method="GET",
            params={
                "project_id": project_id,
                "page": page,
                "size": size,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerJobPage,
                    parse_obj_as(
                        type_=LocalRunnerJobPage,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def list_runners(
        self,
        *,
        project_id: str,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[LocalRunnerPage]:
        """
        List local runners owned by the current user in the workspace

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[LocalRunnerPage]
            Runners list
        """
        _response = self._client_wrapper.httpx_client.request(
            "v1/private/local-runners",
            method="GET",
            params={
                "project_id": project_id,
                "page": page,
                "size": size,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerPage,
                    parse_obj_as(
                        type_=LocalRunnerPage,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def next_job(
        self, runner_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> HttpResponse[LocalRunnerJob]:
        """
        Long-poll for the next pending local runner job

        Parameters
        ----------
        runner_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[LocalRunnerJob]
            Job available
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/{jsonable_encoder(runner_id)}/jobs/next",
            method="POST",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerJob,
                    parse_obj_as(
                        type_=LocalRunnerJob,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return HttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def register_agents(
        self,
        runner_id: str,
        *,
        request: typing.Dict[str, typing.Optional[typing.Any]],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Register or update the local runner's agent list

        Parameters
        ----------
        runner_id : str

        request : typing.Dict[str, typing.Optional[typing.Any]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/{jsonable_encoder(runner_id)}/agents",
            method="PUT",
            json=request,
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    def report_job_result(
        self,
        job_id: str,
        *,
        status: LocalRunnerJobResultRequestStatus,
        result: typing.Optional[JsonNode] = OMIT,
        error: typing.Optional[str] = OMIT,
        trace_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> HttpResponse[None]:
        """
        Report local runner job completion or failure

        Parameters
        ----------
        job_id : str

        status : LocalRunnerJobResultRequestStatus

        result : typing.Optional[JsonNode]

        error : typing.Optional[str]

        trace_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        HttpResponse[None]
        """
        _response = self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/jobs/{jsonable_encoder(job_id)}/results",
            method="POST",
            json={
                "status": status,
                "result": result,
                "error": error,
                "trace_id": trace_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return HttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)


class AsyncRawRunnersClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._client_wrapper = client_wrapper

    async def get_job_logs(
        self,
        job_id: str,
        *,
        offset: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[typing.List[LocalRunnerLogEntry]]:
        """
        Get log entries for a local runner job

        Parameters
        ----------
        job_id : str

        offset : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[typing.List[LocalRunnerLogEntry]]
            Log entries
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/jobs/{jsonable_encoder(job_id)}/logs",
            method="GET",
            params={
                "offset": offset,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    typing.List[LocalRunnerLogEntry],
                    parse_obj_as(
                        type_=typing.List[LocalRunnerLogEntry],  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def append_job_logs(
        self,
        job_id: str,
        *,
        request: typing.Sequence[LocalRunnerLogEntry],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Append log entries for a running local runner job

        Parameters
        ----------
        job_id : str

        request : typing.Sequence[LocalRunnerLogEntry]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/jobs/{jsonable_encoder(job_id)}/logs",
            method="POST",
            json=convert_and_respect_annotation_metadata(
                object_=request, annotation=typing.Sequence[LocalRunnerLogEntry], direction="write"
            ),
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def cancel_job(
        self, job_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[None]:
        """
        Cancel a pending or running local runner job

        Parameters
        ----------
        job_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/jobs/{jsonable_encoder(job_id)}/cancel",
            method="POST",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def connect_runner(
        self, *, pairing_code: str, runner_name: str, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[LocalRunnerConnectResponse]:
        """
        Exchange a pairing code or API key for local runner credentials

        Parameters
        ----------
        pairing_code : str

        runner_name : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[LocalRunnerConnectResponse]
            Runner connected
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/local-runners/connections",
            method="POST",
            json={
                "pairing_code": pairing_code,
                "runner_name": runner_name,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerConnectResponse,
                    parse_obj_as(
                        type_=LocalRunnerConnectResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def create_job(
        self,
        *,
        agent_name: str,
        project_id: str,
        inputs: typing.Optional[JsonNode] = OMIT,
        mask_id: typing.Optional[str] = OMIT,
        metadata: typing.Optional[LocalRunnerJobMetadata] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Create a local runner job and enqueue it for execution

        Parameters
        ----------
        agent_name : str

        project_id : str

        inputs : typing.Optional[JsonNode]

        mask_id : typing.Optional[str]

        metadata : typing.Optional[LocalRunnerJobMetadata]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/local-runners/jobs",
            method="POST",
            json={
                "agent_name": agent_name,
                "inputs": inputs,
                "project_id": project_id,
                "mask_id": mask_id,
                "metadata": convert_and_respect_annotation_metadata(
                    object_=metadata, annotation=LocalRunnerJobMetadata, direction="write"
                ),
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 409:
                raise ConflictError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def generate_pairing_code(
        self, *, project_id: str, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[LocalRunnerPairResponse]:
        """
        Generate a pairing code for a local runner in the current workspace

        Parameters
        ----------
        project_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[LocalRunnerPairResponse]
            Pairing code generated
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/local-runners/pairs",
            method="POST",
            json={
                "project_id": project_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerPairResponse,
                    parse_obj_as(
                        type_=LocalRunnerPairResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_job(
        self, job_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[LocalRunnerJob]:
        """
        Get a single local runner job's status and results

        Parameters
        ----------
        job_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[LocalRunnerJob]
            Job details
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/jobs/{jsonable_encoder(job_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerJob,
                    parse_obj_as(
                        type_=LocalRunnerJob,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def get_runner(
        self, runner_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[LocalRunner]:
        """
        Get a single local runner with its registered agents

        Parameters
        ----------
        runner_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[LocalRunner]
            Runner details
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/{jsonable_encoder(runner_id)}",
            method="GET",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunner,
                    parse_obj_as(
                        type_=LocalRunner,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def heartbeat(
        self, runner_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[LocalRunnerHeartbeatResponse]:
        """
        Refresh local runner heartbeat

        Parameters
        ----------
        runner_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[LocalRunnerHeartbeatResponse]
            Heartbeat response
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/{jsonable_encoder(runner_id)}/heartbeats",
            method="POST",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerHeartbeatResponse,
                    parse_obj_as(
                        type_=LocalRunnerHeartbeatResponse,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 410:
                raise GoneError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        ErrorMessage,
                        parse_obj_as(
                            type_=ErrorMessage,  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def list_jobs(
        self,
        runner_id: str,
        *,
        project_id: typing.Optional[str] = None,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[LocalRunnerJobPage]:
        """
        List jobs for a local runner

        Parameters
        ----------
        runner_id : str

        project_id : typing.Optional[str]

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[LocalRunnerJobPage]
            Jobs list
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/{jsonable_encoder(runner_id)}/jobs",
            method="GET",
            params={
                "project_id": project_id,
                "page": page,
                "size": size,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerJobPage,
                    parse_obj_as(
                        type_=LocalRunnerJobPage,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def list_runners(
        self,
        *,
        project_id: str,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[LocalRunnerPage]:
        """
        List local runners owned by the current user in the workspace

        Parameters
        ----------
        project_id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[LocalRunnerPage]
            Runners list
        """
        _response = await self._client_wrapper.httpx_client.request(
            "v1/private/local-runners",
            method="GET",
            params={
                "project_id": project_id,
                "page": page,
                "size": size,
            },
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerPage,
                    parse_obj_as(
                        type_=LocalRunnerPage,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def next_job(
        self, runner_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> AsyncHttpResponse[LocalRunnerJob]:
        """
        Long-poll for the next pending local runner job

        Parameters
        ----------
        runner_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[LocalRunnerJob]
            Job available
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/{jsonable_encoder(runner_id)}/jobs/next",
            method="POST",
            request_options=request_options,
        )
        try:
            if 200 <= _response.status_code < 300:
                _data = typing.cast(
                    LocalRunnerJob,
                    parse_obj_as(
                        type_=LocalRunnerJob,  # type: ignore
                        object_=_response.json(),
                    ),
                )
                return AsyncHttpResponse(response=_response, data=_data)
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def register_agents(
        self,
        runner_id: str,
        *,
        request: typing.Dict[str, typing.Optional[typing.Any]],
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Register or update the local runner's agent list

        Parameters
        ----------
        runner_id : str

        request : typing.Dict[str, typing.Optional[typing.Any]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/{jsonable_encoder(runner_id)}/agents",
            method="PUT",
            json=request,
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)

    async def report_job_result(
        self,
        job_id: str,
        *,
        status: LocalRunnerJobResultRequestStatus,
        result: typing.Optional[JsonNode] = OMIT,
        error: typing.Optional[str] = OMIT,
        trace_id: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> AsyncHttpResponse[None]:
        """
        Report local runner job completion or failure

        Parameters
        ----------
        job_id : str

        status : LocalRunnerJobResultRequestStatus

        result : typing.Optional[JsonNode]

        error : typing.Optional[str]

        trace_id : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        AsyncHttpResponse[None]
        """
        _response = await self._client_wrapper.httpx_client.request(
            f"v1/private/local-runners/jobs/{jsonable_encoder(job_id)}/results",
            method="POST",
            json={
                "status": status,
                "result": result,
                "error": error,
                "trace_id": trace_id,
            },
            headers={
                "content-type": "application/json",
            },
            request_options=request_options,
            omit=OMIT,
        )
        try:
            if 200 <= _response.status_code < 300:
                return AsyncHttpResponse(response=_response, data=None)
            if _response.status_code == 400:
                raise BadRequestError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            if _response.status_code == 404:
                raise NotFoundError(
                    headers=dict(_response.headers),
                    body=typing.cast(
                        typing.Optional[typing.Any],
                        parse_obj_as(
                            type_=typing.Optional[typing.Any],  # type: ignore
                            object_=_response.json(),
                        ),
                    ),
                )
            _response_json = _response.json()
        except JSONDecodeError:
            raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response.text)
        raise ApiError(status_code=_response.status_code, headers=dict(_response.headers), body=_response_json)



# This file was auto-generated by Fern from our API Definition.

import typing

from ..core.client_wrapper import AsyncClientWrapper, SyncClientWrapper
from ..core.request_options import RequestOptions
from ..types.json_node_write import JsonNodeWrite
from ..types.prompt_detail import PromptDetail
from ..types.prompt_page_public import PromptPagePublic
from ..types.prompt_version_detail import PromptVersionDetail
from ..types.prompt_version_link_public import PromptVersionLinkPublic
from ..types.prompt_version_page_public import PromptVersionPagePublic
from ..types.prompt_version_update import PromptVersionUpdate
from .raw_client import AsyncRawPromptsClient, RawPromptsClient
from .types.create_prompt_version_detail_template_structure import CreatePromptVersionDetailTemplateStructure
from .types.prompt_write_template_structure import PromptWriteTemplateStructure
from .types.prompt_write_type import PromptWriteType

# this is used as the default value for optional parameters
OMIT = typing.cast(typing.Any, ...)


class PromptsClient:
    def __init__(self, *, client_wrapper: SyncClientWrapper):
        self._raw_client = RawPromptsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> RawPromptsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        RawPromptsClient
        """
        return self._raw_client

    def get_prompts(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        project_id: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptPagePublic:
        """
        Get prompts

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        project_id : typing.Optional[str]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptPagePublic
            OK

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.get_prompts()
        """
        _response = self._raw_client.get_prompts(
            page=page,
            size=size,
            name=name,
            project_id=project_id,
            sorting=sorting,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    def create_prompt(
        self,
        *,
        name: str,
        id: typing.Optional[str] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        description: typing.Optional[str] = OMIT,
        template: typing.Optional[str] = OMIT,
        metadata: typing.Optional[JsonNodeWrite] = OMIT,
        change_description: typing.Optional[str] = OMIT,
        type: typing.Optional[PromptWriteType] = OMIT,
        template_structure: typing.Optional[PromptWriteTemplateStructure] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create prompt

        Parameters
        ----------
        name : str

        id : typing.Optional[str]

        project_id : typing.Optional[str]
            Project ID. Takes precedence over project_name when both are provided.

        project_name : typing.Optional[str]
            For project scope, specify either project_id or project_name. If project_name is provided and the project does not exist, it will be created. Ignored when project_id is provided. If neither is provided, the prompt is created at workspace level.

        description : typing.Optional[str]

        template : typing.Optional[str]

        metadata : typing.Optional[JsonNodeWrite]

        change_description : typing.Optional[str]

        type : typing.Optional[PromptWriteType]

        template_structure : typing.Optional[PromptWriteTemplateStructure]
            Template structure type: 'text' or 'chat'. Immutable after creation.

        tags : typing.Optional[typing.Sequence[str]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.create_prompt(name='name', )
        """
        _response = self._raw_client.create_prompt(
            name=name,
            id=id,
            project_id=project_id,
            project_name=project_name,
            description=description,
            template=template,
            metadata=metadata,
            change_description=change_description,
            type=type,
            template_structure=template_structure,
            tags=tags,
            request_options=request_options,
        )
        return _response.data

    def create_prompt_version(
        self,
        *,
        name: str,
        version: PromptVersionDetail,
        template_structure: typing.Optional[CreatePromptVersionDetailTemplateStructure] = OMIT,
        exclude_blueprint_update_for_projects: typing.Optional[typing.Sequence[str]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptVersionDetail:
        """
        Create prompt version

        Parameters
        ----------
        name : str

        version : PromptVersionDetail

        template_structure : typing.Optional[CreatePromptVersionDetailTemplateStructure]
            Template structure for the prompt: 'text' or 'chat'. Note: This field is only used when creating a new prompt. If a prompt with the given name already exists, this field is ignored and the existing prompt's template structure is used. Template structure is immutable after prompt creation.

        exclude_blueprint_update_for_projects : typing.Optional[typing.Sequence[str]]
            Optional set of project IDs to exclude from automatic blueprint creation when this prompt version is committed.

        project_id : typing.Optional[str]
            Project ID. Takes precedence over project_name when both are provided.

        project_name : typing.Optional[str]
            If provided, scopes the prompt to the specified project. Ignored when project_id is provided.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptVersionDetail
            OK

        Examples
        --------
        from Opik import OpikApi
        from Opik import PromptVersionDetail
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.create_prompt_version(name='name', version=PromptVersionDetail(template='template', ), )
        """
        _response = self._raw_client.create_prompt_version(
            name=name,
            version=version,
            template_structure=template_structure,
            exclude_blueprint_update_for_projects=exclude_blueprint_update_for_projects,
            project_id=project_id,
            project_name=project_name,
            request_options=request_options,
        )
        return _response.data

    def update_prompt_versions(
        self,
        *,
        ids: typing.Sequence[str],
        update: PromptVersionUpdate,
        merge_tags: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update one or more prompt versions.

        Note: Prompt versions are immutable by design.
        Only organizational properties, such as tags etc., can be updated.
        Core properties like template and metadata cannot be modified after creation.

        PATCH semantics:
        - non-empty values update the field
        - null values preserve existing field values (no change)
        - empty values explicitly clear the field

        Parameters
        ----------
        ids : typing.Sequence[str]
            IDs of prompt versions to update

        update : PromptVersionUpdate

        merge_tags : typing.Optional[bool]
            Tag merge behavior:
            - true: Add new tags to existing tags (union)
            - false: Replace all existing tags with new tags (default behaviour if not provided)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        from Opik import PromptVersionUpdate
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.update_prompt_versions(ids=['ids'], update=PromptVersionUpdate(), )
        """
        _response = self._raw_client.update_prompt_versions(
            ids=ids, update=update, merge_tags=merge_tags, request_options=request_options
        )
        return _response.data

    def get_prompt_by_id(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> PromptDetail:
        """
        Get prompt by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptDetail
            Prompt resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.get_prompt_by_id(id='id', )
        """
        _response = self._raw_client.get_prompt_by_id(id, request_options=request_options)
        return _response.data

    def update_prompt(
        self,
        id: str,
        *,
        name: str,
        description: typing.Optional[str] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update prompt

        Parameters
        ----------
        id : str

        name : str

        description : typing.Optional[str]

        tags : typing.Optional[typing.Sequence[str]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.update_prompt(id='id', name='name', )
        """
        _response = self._raw_client.update_prompt(
            id, name=name, description=description, tags=tags, request_options=request_options
        )
        return _response.data

    def delete_prompt(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete prompt

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.delete_prompt(id='id', )
        """
        _response = self._raw_client.delete_prompt(id, request_options=request_options)
        return _response.data

    def delete_prompts_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete prompts batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.delete_prompts_batch(ids=['ids'], )
        """
        _response = self._raw_client.delete_prompts_batch(ids=ids, request_options=request_options)
        return _response.data

    def get_prompt_by_commit(
        self, commit: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> PromptDetail:
        """
        Get prompt by commit

        Parameters
        ----------
        commit : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptDetail
            OK

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.get_prompt_by_commit(commit='commit', )
        """
        _response = self._raw_client.get_prompt_by_commit(commit, request_options=request_options)
        return _response.data

    def get_prompt_version_by_id(
        self, version_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> PromptVersionDetail:
        """
        Get prompt version by id

        Parameters
        ----------
        version_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptVersionDetail
            Prompt version resource

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.get_prompt_version_by_id(version_id='versionId', )
        """
        _response = self._raw_client.get_prompt_version_by_id(version_id, request_options=request_options)
        return _response.data

    def get_prompt_versions(
        self,
        id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        search: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptVersionPagePublic:
        """
        Get prompt versions

        Parameters
        ----------
        id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        search : typing.Optional[str]
            Search text to find in template or change description fields

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptVersionPagePublic
            OK

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.get_prompt_versions(id='id', )
        """
        _response = self._raw_client.get_prompt_versions(
            id, page=page, size=size, search=search, sorting=sorting, filters=filters, request_options=request_options
        )
        return _response.data

    def get_prompts_by_commits(
        self, *, commits: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> typing.List[PromptVersionLinkPublic]:
        """
        Get prompts by prompt version commits

        Parameters
        ----------
        commits : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[PromptVersionLinkPublic]
            OK

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.get_prompts_by_commits(commits=['commits'], )
        """
        _response = self._raw_client.get_prompts_by_commits(commits=commits, request_options=request_options)
        return _response.data

    def restore_prompt_version(
        self, prompt_id: str, version_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> PromptVersionDetail:
        """
        Restore a prompt version by creating a new version with the content from the specified version

        Parameters
        ----------
        prompt_id : str

        version_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptVersionDetail
            OK

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.restore_prompt_version(prompt_id='promptId', version_id='versionId', )
        """
        _response = self._raw_client.restore_prompt_version(prompt_id, version_id, request_options=request_options)
        return _response.data

    def retrieve_prompt_version(
        self,
        *,
        name: str,
        commit: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptVersionDetail:
        """
        Retrieve prompt version

        Parameters
        ----------
        name : str

        commit : typing.Optional[str]

        project_name : typing.Optional[str]
            If provided, scopes the search to the specified project

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptVersionDetail
            OK

        Examples
        --------
        from Opik import OpikApi
        client = OpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        client.prompts.retrieve_prompt_version(name='name', )
        """
        _response = self._raw_client.retrieve_prompt_version(
            name=name, commit=commit, project_name=project_name, request_options=request_options
        )
        return _response.data


class AsyncPromptsClient:
    def __init__(self, *, client_wrapper: AsyncClientWrapper):
        self._raw_client = AsyncRawPromptsClient(client_wrapper=client_wrapper)

    @property
    def with_raw_response(self) -> AsyncRawPromptsClient:
        """
        Retrieves a raw implementation of this client that returns raw responses.

        Returns
        -------
        AsyncRawPromptsClient
        """
        return self._raw_client

    async def get_prompts(
        self,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        name: typing.Optional[str] = None,
        project_id: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptPagePublic:
        """
        Get prompts

        Parameters
        ----------
        page : typing.Optional[int]

        size : typing.Optional[int]

        name : typing.Optional[str]

        project_id : typing.Optional[str]

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptPagePublic
            OK

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.get_prompts()
        asyncio.run(main())
        """
        _response = await self._raw_client.get_prompts(
            page=page,
            size=size,
            name=name,
            project_id=project_id,
            sorting=sorting,
            filters=filters,
            request_options=request_options,
        )
        return _response.data

    async def create_prompt(
        self,
        *,
        name: str,
        id: typing.Optional[str] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        description: typing.Optional[str] = OMIT,
        template: typing.Optional[str] = OMIT,
        metadata: typing.Optional[JsonNodeWrite] = OMIT,
        change_description: typing.Optional[str] = OMIT,
        type: typing.Optional[PromptWriteType] = OMIT,
        template_structure: typing.Optional[PromptWriteTemplateStructure] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Create prompt

        Parameters
        ----------
        name : str

        id : typing.Optional[str]

        project_id : typing.Optional[str]
            Project ID. Takes precedence over project_name when both are provided.

        project_name : typing.Optional[str]
            For project scope, specify either project_id or project_name. If project_name is provided and the project does not exist, it will be created. Ignored when project_id is provided. If neither is provided, the prompt is created at workspace level.

        description : typing.Optional[str]

        template : typing.Optional[str]

        metadata : typing.Optional[JsonNodeWrite]

        change_description : typing.Optional[str]

        type : typing.Optional[PromptWriteType]

        template_structure : typing.Optional[PromptWriteTemplateStructure]
            Template structure type: 'text' or 'chat'. Immutable after creation.

        tags : typing.Optional[typing.Sequence[str]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.create_prompt(name='name', )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_prompt(
            name=name,
            id=id,
            project_id=project_id,
            project_name=project_name,
            description=description,
            template=template,
            metadata=metadata,
            change_description=change_description,
            type=type,
            template_structure=template_structure,
            tags=tags,
            request_options=request_options,
        )
        return _response.data

    async def create_prompt_version(
        self,
        *,
        name: str,
        version: PromptVersionDetail,
        template_structure: typing.Optional[CreatePromptVersionDetailTemplateStructure] = OMIT,
        exclude_blueprint_update_for_projects: typing.Optional[typing.Sequence[str]] = OMIT,
        project_id: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptVersionDetail:
        """
        Create prompt version

        Parameters
        ----------
        name : str

        version : PromptVersionDetail

        template_structure : typing.Optional[CreatePromptVersionDetailTemplateStructure]
            Template structure for the prompt: 'text' or 'chat'. Note: This field is only used when creating a new prompt. If a prompt with the given name already exists, this field is ignored and the existing prompt's template structure is used. Template structure is immutable after prompt creation.

        exclude_blueprint_update_for_projects : typing.Optional[typing.Sequence[str]]
            Optional set of project IDs to exclude from automatic blueprint creation when this prompt version is committed.

        project_id : typing.Optional[str]
            Project ID. Takes precedence over project_name when both are provided.

        project_name : typing.Optional[str]
            If provided, scopes the prompt to the specified project. Ignored when project_id is provided.

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptVersionDetail
            OK

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import PromptVersionDetail
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.create_prompt_version(name='name', version=PromptVersionDetail(template='template', ), )
        asyncio.run(main())
        """
        _response = await self._raw_client.create_prompt_version(
            name=name,
            version=version,
            template_structure=template_structure,
            exclude_blueprint_update_for_projects=exclude_blueprint_update_for_projects,
            project_id=project_id,
            project_name=project_name,
            request_options=request_options,
        )
        return _response.data

    async def update_prompt_versions(
        self,
        *,
        ids: typing.Sequence[str],
        update: PromptVersionUpdate,
        merge_tags: typing.Optional[bool] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update one or more prompt versions.

        Note: Prompt versions are immutable by design.
        Only organizational properties, such as tags etc., can be updated.
        Core properties like template and metadata cannot be modified after creation.

        PATCH semantics:
        - non-empty values update the field
        - null values preserve existing field values (no change)
        - empty values explicitly clear the field

        Parameters
        ----------
        ids : typing.Sequence[str]
            IDs of prompt versions to update

        update : PromptVersionUpdate

        merge_tags : typing.Optional[bool]
            Tag merge behavior:
            - true: Add new tags to existing tags (union)
            - false: Replace all existing tags with new tags (default behaviour if not provided)

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        from Opik import PromptVersionUpdate
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.update_prompt_versions(ids=['ids'], update=PromptVersionUpdate(), )
        asyncio.run(main())
        """
        _response = await self._raw_client.update_prompt_versions(
            ids=ids, update=update, merge_tags=merge_tags, request_options=request_options
        )
        return _response.data

    async def get_prompt_by_id(
        self, id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> PromptDetail:
        """
        Get prompt by id

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptDetail
            Prompt resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.get_prompt_by_id(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_prompt_by_id(id, request_options=request_options)
        return _response.data

    async def update_prompt(
        self,
        id: str,
        *,
        name: str,
        description: typing.Optional[str] = OMIT,
        tags: typing.Optional[typing.Sequence[str]] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> None:
        """
        Update prompt

        Parameters
        ----------
        id : str

        name : str

        description : typing.Optional[str]

        tags : typing.Optional[typing.Sequence[str]]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.update_prompt(id='id', name='name', )
        asyncio.run(main())
        """
        _response = await self._raw_client.update_prompt(
            id, name=name, description=description, tags=tags, request_options=request_options
        )
        return _response.data

    async def delete_prompt(self, id: str, *, request_options: typing.Optional[RequestOptions] = None) -> None:
        """
        Delete prompt

        Parameters
        ----------
        id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.delete_prompt(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_prompt(id, request_options=request_options)
        return _response.data

    async def delete_prompts_batch(
        self, *, ids: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> None:
        """
        Delete prompts batch

        Parameters
        ----------
        ids : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        None

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.delete_prompts_batch(ids=['ids'], )
        asyncio.run(main())
        """
        _response = await self._raw_client.delete_prompts_batch(ids=ids, request_options=request_options)
        return _response.data

    async def get_prompt_by_commit(
        self, commit: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> PromptDetail:
        """
        Get prompt by commit

        Parameters
        ----------
        commit : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptDetail
            OK

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.get_prompt_by_commit(commit='commit', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_prompt_by_commit(commit, request_options=request_options)
        return _response.data

    async def get_prompt_version_by_id(
        self, version_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> PromptVersionDetail:
        """
        Get prompt version by id

        Parameters
        ----------
        version_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptVersionDetail
            Prompt version resource

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.get_prompt_version_by_id(version_id='versionId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_prompt_version_by_id(version_id, request_options=request_options)
        return _response.data

    async def get_prompt_versions(
        self,
        id: str,
        *,
        page: typing.Optional[int] = None,
        size: typing.Optional[int] = None,
        search: typing.Optional[str] = None,
        sorting: typing.Optional[str] = None,
        filters: typing.Optional[str] = None,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptVersionPagePublic:
        """
        Get prompt versions

        Parameters
        ----------
        id : str

        page : typing.Optional[int]

        size : typing.Optional[int]

        search : typing.Optional[str]
            Search text to find in template or change description fields

        sorting : typing.Optional[str]

        filters : typing.Optional[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptVersionPagePublic
            OK

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.get_prompt_versions(id='id', )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_prompt_versions(
            id, page=page, size=size, search=search, sorting=sorting, filters=filters, request_options=request_options
        )
        return _response.data

    async def get_prompts_by_commits(
        self, *, commits: typing.Sequence[str], request_options: typing.Optional[RequestOptions] = None
    ) -> typing.List[PromptVersionLinkPublic]:
        """
        Get prompts by prompt version commits

        Parameters
        ----------
        commits : typing.Sequence[str]

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        typing.List[PromptVersionLinkPublic]
            OK

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.get_prompts_by_commits(commits=['commits'], )
        asyncio.run(main())
        """
        _response = await self._raw_client.get_prompts_by_commits(commits=commits, request_options=request_options)
        return _response.data

    async def restore_prompt_version(
        self, prompt_id: str, version_id: str, *, request_options: typing.Optional[RequestOptions] = None
    ) -> PromptVersionDetail:
        """
        Restore a prompt version by creating a new version with the content from the specified version

        Parameters
        ----------
        prompt_id : str

        version_id : str

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptVersionDetail
            OK

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.restore_prompt_version(prompt_id='promptId', version_id='versionId', )
        asyncio.run(main())
        """
        _response = await self._raw_client.restore_prompt_version(
            prompt_id, version_id, request_options=request_options
        )
        return _response.data

    async def retrieve_prompt_version(
        self,
        *,
        name: str,
        commit: typing.Optional[str] = OMIT,
        project_name: typing.Optional[str] = OMIT,
        request_options: typing.Optional[RequestOptions] = None,
    ) -> PromptVersionDetail:
        """
        Retrieve prompt version

        Parameters
        ----------
        name : str

        commit : typing.Optional[str]

        project_name : typing.Optional[str]
            If provided, scopes the search to the specified project

        request_options : typing.Optional[RequestOptions]
            Request-specific configuration.

        Returns
        -------
        PromptVersionDetail
            OK

        Examples
        --------
        from Opik import AsyncOpikApi
        import asyncio
        client = AsyncOpikApi(api_key="YOUR_API_KEY", workspace_name="YOUR_WORKSPACE_NAME", )
        async def main() -> None:
            await client.prompts.retrieve_prompt_version(name='name', )
        asyncio.run(main())
        """
        _response = await self._raw_client.retrieve_prompt_version(
            name=name, commit=commit, project_name=project_name, request_options=request_options
        )
        return _response.data



import abc
from typing import Any, Optional

from . import attachment


class AttachmentDecoder(abc.ABC):
    """
    Abstract base class for decoding file attachments.

    This class serves as an interface for decoding raw attachment data into
    an `Attachment` object. Implementing classes should define the specific
    logic to handle various attachment decoding formats.
    """

    @abc.abstractmethod
    def decode(self, raw_data: str, **kwargs: Any) -> Optional[attachment.Attachment]:
        pass



# This file was auto-generated by Fern from our API Definition.

import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .agent_blueprint_write import AgentBlueprintWrite


class AgentConfigCreateWrite(UniversalBaseModel):
    project_id: typing.Optional[str] = pydantic.Field(default=None)
    """
    Project ID. Either project_id or project_name must be provided
    """

    project_name: typing.Optional[str] = pydantic.Field(default=None)
    """
    Project name. Either project_id or project_name must be provided
    """

    id: typing.Optional[str] = pydantic.Field(default=None)
    """
    Agent config ID. Generated automatically if not provided
    """

    blueprint: AgentBlueprintWrite

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import warnings
from contextlib import contextmanager
from typing import Any, Iterator, List, Optional, Tuple, Union

from opik.exceptions import MetricComputationError
from opik.evaluation.metrics import base_metric, score_result

try:
    from nltk.translate import bleu_score as nltk_bleu_score
except ImportError:
    nltk_bleu_score = None


class BaseBLEU(base_metric.BaseMetric):
    """
    Base class containing shared BLEU logic, such as handling n-grams, smoothing,
    and weights initialization. This class is not intended to be used directly.

    References:
      - BLEU: Papineni et al., "BLEU: a Method for Automatic Evaluation of Machine Translation" (ACL 2002)
        https://aclanthology.org/P02-1040/
      - NLTK BLEU documentation
        https://www.nltk.org/api/nltk.translate.bleu_score.html
      - Hugging Face Evaluate: BLEU metric overview
        https://huggingface.co/spaces/evaluate-metric/bleu

    Args:
        name: The name of the metric (e.g. "sentence_bleu_metric" or "corpus_bleu_metric").
        track: Whether to track the metric (depends on your system).
        n_grams: Up to which n-gram order to use (1 through n_grams).
        smoothing_method: One of NLTK's SmoothingFunction methods (e.g. "method0", "method1", etc.).
        weights: Optional custom weights for n-gram orders. Must sum to 1.0. If None,
                 defaults to uniform distribution across `n_grams`.
        project_name: Optional project name to track the metric in for the cases when
            there are no parent span/trace to inherit project name from.
    """

    def __init__(
        self,
        name: str,
        track: bool,
        n_grams: int,
        smoothing_method: str,
        weights: Optional[List[float]],
        project_name: Optional[str],
    ):
        super().__init__(name=name, track=track, project_name=project_name)

        if nltk_bleu_score is None:
            raise ImportError(
                "`nltk` library is required for BLEU score calculation. "
                "Install via `pip install nltk`."
            )

        self.n_grams = n_grams
        self.smoothing_method = smoothing_method

        if weights is None:
            self.weights = [1.0 / n_grams] * n_grams
        else:
            if len(weights) != n_grams:
                raise ValueError(
                    f"Length of weights ({len(weights)}) != n_grams ({n_grams})."
                )
            if abs(sum(weights) - 1.0) > 1e-6:
                raise ValueError("Weights must sum to 1.0")
            self.weights = weights

        self._smoother = nltk_bleu_score.SmoothingFunction()

    def _get_smoothing_func(self) -> "nltk_bleu_score.SmoothingFunction":
        return getattr(self._smoother, self.smoothing_method, self._smoother.method0)

    def _truncate_weights(self, max_len: int) -> Tuple[float, ...]:
        used_order = min(self.n_grams, max_len)
        used_weights = self.weights[:used_order]
        total = sum(used_weights) or 1.0
        normalized = [w / total for w in used_weights]
        return tuple(normalized)


@contextmanager
def _suppress_bleu_warnings() -> Iterator[None]:
    with warnings.catch_warnings():
        warnings.filterwarnings(
            "ignore",
            message=r"The hypothesis contains 0 counts of 2-gram overlaps\.",
            category=UserWarning,
            module="nltk\\.translate\\.bleu_score",
        )
        yield


class SentenceBLEU(BaseBLEU):
    """
    Computes sentence-level BLEU for a single candidate string vs. one or more references.

    Example:
        >>> from opik.evaluation.metrics.heuristics.bleu import SentenceBLEU
        >>> metric = SentenceBLEU(n_grams=4, smoothing_method="method1")
        >>> result = metric.score("the cat is on the mat", "the cat is on the mat")
        >>> print(result.value)
        1.0
    """

    def __init__(
        self,
        name: str = "sentence_bleu_metric",
        track: bool = True,
        n_grams: int = 4,
        smoothing_method: str = "method1",
        weights: Optional[List[float]] = None,
        project_name: Optional[str] = None,
    ):
        super().__init__(
            name=name,
            track=track,
            n_grams=n_grams,
            smoothing_method=smoothing_method,
            weights=weights,
            project_name=project_name,
        )

    def score(
        self,
        output: str,
        reference: Union[str, List[str]],
        **ignored_kwargs: Any,
    ) -> score_result.ScoreResult:
        """
        Calculate sentence-level BLEU for one candidate vs. one or more references.

        Args:
            output: A single candidate string.
            reference: Either a single reference string or a list of reference strings.

        Returns:
            A `ScoreResult` with:
              - `value`: The sentence-level BLEU score (float).
              - `name`: The metric name.
              - `reason`: A short explanation (e.g. "Sentence-level BLEU...").

        Raises:
            MetricComputationError:
                - If the candidate or any reference is empty.
        """
        if not output.strip():
            raise MetricComputationError("Candidate is empty (single-sentence BLEU).")

        # Convert references to tokenized lists
        if isinstance(reference, str):
            if not reference.strip():
                raise MetricComputationError(
                    "Reference is empty (single-sentence BLEU)."
                )
            ref_lists = [reference.lower().split()]
        else:
            # List of reference strings
            ref_lists = []
            for ref_str in reference:
                if not ref_str.strip():
                    raise MetricComputationError(
                        "Encountered empty reference (single-sentence BLEU)."
                    )
                ref_lists.append(ref_str.lower().split())

        candidate_tokens = output.lower().split()
        used_weights = self._truncate_weights(len(candidate_tokens))
        smoothing_func = self._get_smoothing_func()

        try:
            with _suppress_bleu_warnings():
                bleu_val = nltk_bleu_score.sentence_bleu(
                    ref_lists,
                    candidate_tokens,
                    weights=used_weights,
                    smoothing_function=smoothing_func,
                )
        except ZeroDivisionError:
            bleu_val = 0.0

        return score_result.ScoreResult(
            value=bleu_val,
            name=self.name,
            reason=(
                f"Sentence-level BLEU (nltk, method={self.smoothing_method}): {bleu_val:.4f}"
            ),
        )


class CorpusBLEU(BaseBLEU):
    """
    Computes corpus-level BLEU for multiple candidate strings vs. matching references.

    Each element in `output` corresponds to one candidate. The parallel `reference`
    element can be either a single string or a list of reference strings for that candidate.

    Example:
        >>> from opik.evaluation.metrics.heuristics.bleu import CorpusBLEU
        >>> metric = CorpusBLEU(n_grams=4, smoothing_method="method1")
        >>> outputs = ["the cat is on the mat", "there is a cat here"]
        >>> references = [
        ...     "the cat is on the mat",
        ...     ["there is a cat here", "there is cat here"]
        ... ]
        >>> result = metric.score(outputs, references)
        >>> print(result.value)
    """

    def __init__(
        self,
        name: str = "corpus_bleu_metric",
        track: bool = True,
        n_grams: int = 4,
        smoothing_method: str = "method1",
        weights: Optional[List[float]] = None,
        project_name: Optional[str] = None,
    ):
        super().__init__(
            name=name,
            track=track,
            n_grams=n_grams,
            smoothing_method=smoothing_method,
            weights=weights,
            project_name=project_name,
        )

    def score(
        self,
        output: List[str],
        reference: List[Union[str, List[str]]],
        **ignored_kwargs: Any,
    ) -> score_result.ScoreResult:
        """
        Calculate corpus-level BLEU for multiple candidates.

        Args:
            output: A list of candidate strings (one per sample).
            reference: A list of references, each parallel to `output`. Each reference
                       item can be a single string or a list of strings.

        Returns:
            A `ScoreResult` with:
              - `value`: The corpus-level BLEU score (float).
              - `name`: The metric name.
              - `reason`: A short explanation (e.g. "Corpus-level BLEU...").

        Raises:
            MetricComputationError:
                - If a candidate or reference is empty.
                - If the number of candidates does not match the number of references.
        """
        if len(output) != len(reference):
            raise MetricComputationError(
                "Mismatch: number of candidates != number of references (corpus BLEU)."
            )

        all_candidates: List[List[str]] = []
        all_references: List[List[List[str]]] = []

        for candidate_str, ref_item in zip(output, reference):
            if not candidate_str.strip():
                raise MetricComputationError("Candidate is empty (corpus BLEU).")

            candidate_tokens = candidate_str.lower().split()

            if isinstance(ref_item, str):
                # single reference
                if not ref_item.strip():
                    raise MetricComputationError("Reference is empty (corpus BLEU).")
                ref_lists = [ref_item.lower().split()]
            else:
                # multiple references
                ref_lists = []
                for r_line in ref_item:
                    if not r_line.strip():
                        raise MetricComputationError(
                            "Encountered empty reference (corpus BLEU)."
                        )
                    ref_lists.append(r_line.lower().split())

            all_candidates.append(candidate_tokens)
            all_references.append(ref_lists)

        max_candidate_len = max(len(cand) for cand in all_candidates)
        used_weights = self._truncate_weights(max_candidate_len)
        smoothing_func = self._get_smoothing_func()

        try:
            with _suppress_bleu_warnings():
                bleu_val = nltk_bleu_score.corpus_bleu(
                    all_references,
                    all_candidates,
                    weights=used_weights,
                    smoothing_function=smoothing_func,
                )
        except ZeroDivisionError:
            bleu_val = 0.0

        return score_result.ScoreResult(
            value=bleu_val,
            name=self.name,
            reason=(
                f"Corpus-level BLEU (nltk, method={self.smoothing_method}): {bleu_val:.4f}"
            ),
        )



# This file was auto-generated by Fern from our API Definition.

import typing

AgentConfigValueHistoryType = typing.Union[
    typing.Literal["string", "integer", "float", "boolean", "prompt", "prompt_commit"], typing.Any
]



from __future__ import annotations

import json
from typing import Callable, List, Optional, Type, TypeVar

from opik.rest_api import OpikApi
from opik.rest_api.types import AnnotationQueuePublic
import opik.exceptions as exceptions
from . import annotation_queue
from ...rest_api.core.api_error import ApiError

QueueT = TypeVar(
    "QueueT",
    annotation_queue.TracesAnnotationQueue,
    annotation_queue.ThreadsAnnotationQueue,
)


def _create_queue_instance(
    queue_data: AnnotationQueuePublic,
    rest_client: OpikApi,
    queue_class: Type[QueueT],
) -> QueueT:
    """Helper to create an annotation queue instance from API response data."""
    return queue_class(
        id=queue_data.id or "",
        name=queue_data.name,
        project_id=queue_data.project_id,
        rest_client=rest_client,
        description=queue_data.description,
        instructions=queue_data.instructions,
        comments_enabled=queue_data.comments_enabled,
        feedback_definition_names=list(queue_data.feedback_definition_names)
        if queue_data.feedback_definition_names
        else None,
        items_count=queue_data.items_count,
    )


def _get_annotation_queues_by_scope(
    rest_client: OpikApi,
    queue_class: Type[QueueT],
    scope_filter: Callable[[Optional[str]], bool],
    project_id: Optional[str] = None,
    max_results: int = 1000,
) -> List[QueueT]:
    """Helper to fetch annotation queues filtered by scope."""
    page_size = 100
    queues: List[QueueT] = []

    filters: Optional[str] = None
    if project_id is not None:
        filters = json.dumps(
            [
                {
                    "field": "project_id",
                    "type": "string",
                    "operator": "=",
                    "value": project_id,
                }
            ]
        )

    page = 1
    while len(queues) < max_results:
        page_queues = rest_client.annotation_queues.find_annotation_queues(
            page=page,
            size=page_size,
            filters=filters,
        )

        if page_queues.content is None or len(page_queues.content) == 0:
            break

        for queue_data in page_queues.content:
            if len(queues) >= max_results:
                break
            if scope_filter(queue_data.scope):
                queues.append(
                    _create_queue_instance(queue_data, rest_client, queue_class)
                )

        page += 1

    return queues


def _get_annotation_queue_by_id_with_scope(
    rest_client: OpikApi,
    queue_id: str,
    queue_class: Type[QueueT],
    scope_check: Callable[[Optional[str]], bool],
    scope_name: str,
) -> QueueT:
    """Helper to fetch an annotation queue by ID with scope validation."""
    try:
        queue_data = rest_client.annotation_queues.get_annotation_queue_by_id(queue_id)
    except ApiError as e:
        if e.status_code == 404:
            raise exceptions.OpikException(
                f"Annotation queue with id '{queue_id}' not found."
            ) from e
        raise

    if not scope_check(queue_data.scope):
        actual_scope = queue_data.scope or "trace"
        raise exceptions.OpikException(
            f"Annotation queue with id '{queue_id}' is not a {scope_name} queue (scope: {actual_scope})."
        )

    return _create_queue_instance(queue_data, rest_client, queue_class)


def get_traces_annotation_queues(
    rest_client: OpikApi,
    project_id: Optional[str] = None,
    max_results: int = 1000,
) -> List[annotation_queue.TracesAnnotationQueue]:
    """
    Fetch trace annotation queues with optional project filtering.

    Args:
        rest_client: The REST API client.
        project_id: Optional project ID to filter queues.
        max_results: Maximum number of queues to return.

    Returns:
        A list of TracesAnnotationQueue objects.
    """
    return _get_annotation_queues_by_scope(
        rest_client=rest_client,
        queue_class=annotation_queue.TracesAnnotationQueue,
        scope_filter=lambda s: s == "trace",
        project_id=project_id,
        max_results=max_results,
    )


def get_threads_annotation_queues(
    rest_client: OpikApi,
    project_id: Optional[str] = None,
    max_results: int = 1000,
) -> List[annotation_queue.ThreadsAnnotationQueue]:
    """
    Fetch thread annotation queues with optional project filtering.

    Args:
        rest_client: The REST API client.
        project_id: Optional project ID to filter queues.
        max_results: Maximum number of queues to return.

    Returns:
        A list of ThreadsAnnotationQueue objects.
    """
    return _get_annotation_queues_by_scope(
        rest_client=rest_client,
        queue_class=annotation_queue.ThreadsAnnotationQueue,
        scope_filter=lambda s: s == "thread",
        project_id=project_id,
        max_results=max_results,
    )


def get_traces_annotation_queue_by_id(
    rest_client: OpikApi,
    queue_id: str,
) -> annotation_queue.TracesAnnotationQueue:
    """
    Fetch a trace annotation queue by its ID.

    Args:
        rest_client: The REST API client.
        queue_id: The ID of the annotation queue.

    Returns:
        A TracesAnnotationQueue object.

    Raises:
        OpikException: If the queue is not found or is not a trace queue.
    """
    return _get_annotation_queue_by_id_with_scope(
        rest_client=rest_client,
        queue_id=queue_id,
        queue_class=annotation_queue.TracesAnnotationQueue,
        scope_check=lambda s: s == "trace",
        scope_name="traces",
    )


def get_threads_annotation_queue_by_id(
    rest_client: OpikApi,
    queue_id: str,
) -> annotation_queue.ThreadsAnnotationQueue:
    """
    Fetch a thread annotation queue by its ID.

    Args:
        rest_client: The REST API client.
        queue_id: The ID of the annotation queue.

    Returns:
        A ThreadsAnnotationQueue object.

    Raises:
        OpikException: If the queue is not found or is not a thread queue.
    """
    return _get_annotation_queue_by_id_with_scope(
        rest_client=rest_client,
        queue_id=queue_id,
        queue_class=annotation_queue.ThreadsAnnotationQueue,
        scope_check=lambda s: s == "thread",
        scope_name="threads",
    )



"""Healthcheck command for Opik CLI."""

from typing import Optional

import click

from opik.cli.healthcheck.smoke_test import run_smoke_test
from opik.cli.healthcheck import (
    check_user_permissions,
    standard_check,
    rich_representation,
)


@click.command(context_settings={"ignore_unknown_options": True})
@click.option(
    "--show-installed-packages",
    is_flag=True,
    default=False,
    help="Print the list of installed packages to the console.",
)
@click.option(
    "--smoke-test",
    type=str,
    default=None,
    help="Run a smoke test to verify Opik integration is working correctly. Requires WORKSPACE value.",
    metavar="WORKSPACE",
)
@click.option(
    "--project-name",
    type=str,
    default=None,
    help="Project name for the smoke test (only used when --smoke-test is provided). Defaults to 'smoke-test-project'.",
)
@click.option(
    "--check-permissions",
    type=str,
    default=None,
    help="Allows to check user permissions in specific workspace using API key either provided using --api-key or from config file. Requires WORKSPACE value.",
    metavar="WORKSPACE",
)
@click.pass_context
def healthcheck(
    ctx: click.Context,
    show_installed_packages: bool,
    smoke_test: Optional[str],
    project_name: Optional[str],
    check_permissions: Optional[str],
) -> None:
    """
    Performs a health check of the application, including validation of configuration,
    verification of library installations, and checking the availability of the backend workspace.
    Prints all relevant information to assist in debugging and diagnostics.

    When --smoke-test is used, runs a smoke test to verify Opik integration is working correctly.
    The smoke test performs a basic sanity test by:
    - Creating a trace with random data
    - Adding a tracked LLM completion span
    - Attaching a dynamically generated test image
    - Verifying data is sent to Opik

    Examples:

      Run standard healthcheck:

        opik healthcheck

      Run healthcheck with smoke test:

        opik healthcheck --smoke-test my-workspace

      Run smoke test with custom project name:

        opik healthcheck --smoke-test my-workspace --project-name my-test-project

      Run healthcheck with specific workspace permissions check:

        opik healthcheck --check-permissions my-workspace
    """
    # Always run the standard healthcheck
    standard_check.run(show_installed_packages)

    # Validate that --project-name is only used with --smoke-test
    if project_name is not None and smoke_test is None:
        raise click.BadParameter(
            "--project-name can only be used with --smoke-test",
            param_hint="--project-name",
        )

    if smoke_test is not None:
        # Use default project name if smoke_test is provided but project_name is not
        if project_name is None:
            project_name = "smoke-test-project"

        # Validate that workspace is not empty
        _check_workspace(smoke_test, "--smoke-test")

        # Get API key and debug flag from context
        api_key = ctx.obj.get("api_key") if ctx.obj else None
        debug = ctx.obj.get("debug", False) if ctx.obj else False

        try:
            run_smoke_test(
                workspace=smoke_test,
                project_name=project_name,
                api_key=api_key,
                debug=debug,
            )
        except Exception as e:
            click.echo(f"Smoke test failed: {e}", err=True)
            if debug:
                import traceback

                click.echo("Traceback:", err=True)
                click.echo(traceback.format_exc(), err=True)

    if check_permissions is not None:
        # Validate that the workspace is not empty
        _check_workspace(check_permissions, "--check-permissions")

        api_key = ctx.obj.get("api_key") if ctx.obj else None
        check_user_permissions.run(api_key=api_key, workspace=check_permissions)

    rich_representation.print_header("healthcheck completed")


def _check_workspace(workspace: str, param_hint: str) -> None:
    if workspace == "":
        click.echo("---------------------------------------------------------")
        raise click.BadParameter(
            f"{param_hint} requires a non-empty workspace name", param_hint=param_hint
        )



# This file was auto-generated by Fern from our API Definition.

import datetime as dt
import typing

import pydantic
from ..core.pydantic_utilities import IS_PYDANTIC_V2, UniversalBaseModel
from .error_info_experiment_item_bulk_write_view import ErrorInfoExperimentItemBulkWriteView
from .json_list_string_experiment_item_bulk_write_view import JsonListStringExperimentItemBulkWriteView
from .span_experiment_item_bulk_write_view_source import SpanExperimentItemBulkWriteViewSource
from .span_experiment_item_bulk_write_view_type import SpanExperimentItemBulkWriteViewType


class SpanExperimentItemBulkWriteView(UniversalBaseModel):
    id: typing.Optional[str] = None
    parent_span_id: typing.Optional[str] = None
    name: typing.Optional[str] = None
    type: typing.Optional[SpanExperimentItemBulkWriteViewType] = None
    start_time: dt.datetime
    end_time: typing.Optional[dt.datetime] = None
    input: typing.Optional[JsonListStringExperimentItemBulkWriteView] = None
    output: typing.Optional[JsonListStringExperimentItemBulkWriteView] = None
    metadata: typing.Optional[JsonListStringExperimentItemBulkWriteView] = None
    model: typing.Optional[str] = None
    provider: typing.Optional[str] = None
    tags: typing.Optional[typing.List[str]] = None
    usage: typing.Optional[typing.Dict[str, int]] = None
    error_info: typing.Optional[ErrorInfoExperimentItemBulkWriteView] = None
    last_updated_at: typing.Optional[dt.datetime] = None
    total_estimated_cost: typing.Optional[float] = None
    total_estimated_cost_version: typing.Optional[str] = None
    ttft: typing.Optional[float] = pydantic.Field(default=None)
    """
    Time to first token in milliseconds
    """

    source: typing.Optional[SpanExperimentItemBulkWriteViewSource] = None

    if IS_PYDANTIC_V2:
        model_config: typing.ClassVar[pydantic.ConfigDict] = pydantic.ConfigDict(extra="allow", frozen=True)  # type: ignore # Pydantic v2
    else:

        class Config:
            frozen = True
            smart_union = True
            extra = pydantic.Extra.allow



import logging
from contextlib import contextmanager
from typing import Generator, Any

from opik import context_storage
from opik.api_objects import span, trace
from opik.types import DistributedTraceHeadersDict

LOGGER = logging.getLogger(__name__)


@contextmanager
def distributed_headers(
    headers: DistributedTraceHeadersDict,
) -> Generator[None, Any, None]:
    """
    Context manager for managing distributed tracing headers.

    This context manager is used to handle distributed tracing headers in a
    structured manner. It ensures root span creation, error logging during user
    script execution, and cleanup of root span data after use.

    Args:
        headers: Distributed tracing headers used for root span creation.
    """
    if not headers or not headers.get("opik_trace_id"):
        LOGGER.warning(
            "Empty distributed headers provided. Skipping setting distributed headers."
        )
        yield
        return

    opik_context_storage = context_storage.get_current_context_instance()

    trace_id = headers["opik_trace_id"]
    trace_data = trace.TraceData(id=trace_id)
    opik_context_storage.set_trace_data(trace_data)

    parent_span_id = headers.get("opik_parent_span_id")
    if parent_span_id is not None:
        # this is a fake span that imitates a span created somewhere,
        # so using opik_parent_span_id as span_id
        span_data = span.SpanData(
            id=parent_span_id,
            trace_id=trace_id,
        )
        context_storage.add_span_data(span_data)
    else:
        span_data = None

    try:
        yield
    except Exception as exception:
        LOGGER.error(
            "Error in user's script while executing distributed headers context manager: %s",
            str(exception),
            exc_info=True,
        )

        raise
    finally:
        # Clean up fake trace/span data from context
        opik_context_storage.pop_trace_data(ensure_id=trace_id)
        if span_data is not None:
            opik_context_storage.pop_span_data(ensure_id=span_data.id)



"""
This file contains the OQL parser and validator. It is currently limited in scope to only support
simple filters without "or" operators.
"""

import json
from abc import ABC, abstractmethod
from typing import Any, Dict, Optional, Tuple, List

STRING_OPERATORS = [
    "=",
    "!=",
    "contains",
    "not_contains",
    "starts_with",
    "ends_with",
    ">",
    "<",
]
DATE_TIME_OPERATORS = ["=", "!=", ">", ">=", "<", "<="]
NUMBER_OPERATORS = ["=", "!=", ">", ">=", "<", "<="]
FEEDBACK_SCORES_OPERATORS = [
    "=",
    "!=",
    ">",
    ">=",
    "<",
    "<=",
    "is_empty",
    "is_not_empty",
]
LIST_OPERATORS = [
    "=",
    "!=",
    "contains",
    "not_contains",
    "is_empty",
    "is_not_empty",
]
DICTIONARY_OPERATORS = [
    "=",
    "!=",
    "contains",
    "not_contains",
    "starts_with",
    "ends_with",
    ">",
    ">=",
    "<",
    "<=",
]


class OQLConfig(ABC):
    """Abstract base class for OQL configuration."""

    @property
    @abstractmethod
    def columns(self) -> Dict[str, str]:
        """Return the supported columns and their types."""
        pass

    @property
    @abstractmethod
    def supported_operators(self) -> Dict[str, List[str]]:
        """Return the supported operators for each column."""
        pass

    @property
    def dictionary_fields(self) -> List[str]:
        """Return fields that support nested key access via dot notation."""
        return ["usage", "feedback_scores", "metadata"]


class TraceOQLConfig(OQLConfig):
    """OQL configuration for trace filtering.

    Based on backend's TraceField enum.
    See: apps/opik-backend/src/main/java/com/comet/opik/api/filter/TraceField.java
    """

    @property
    def columns(self) -> Dict[str, str]:
        return {
            "id": "string",
            "name": "string",
            "start_time": "date_time",
            "end_time": "date_time",
            "input": "string",
            "output": "string",
            "input_json": "dictionary",
            "output_json": "dictionary",
            "metadata": "dictionary",
            "total_estimated_cost": "number",
            "llm_span_count": "number",
            "tags": "list",
            "usage.total_tokens": "number",
            "usage.prompt_tokens": "number",
            "usage.completion_tokens": "number",
            "feedback_scores": "feedback_scores_number",
            "span_feedback_scores": "feedback_scores_number",
            "duration": "number",
            "thread_id": "string",
            "guardrails": "string",
            "error_info": "error_container",
            "created_at": "date_time",
            "last_updated_at": "date_time",
            "annotation_queue_ids": "list",
            "experiment_id": "string",
        }

    @property
    def supported_operators(self) -> Dict[str, List[str]]:
        return {
            "id": STRING_OPERATORS,
            "name": STRING_OPERATORS,
            "input": STRING_OPERATORS,
            "output": STRING_OPERATORS,
            "thread_id": STRING_OPERATORS,
            "guardrails": STRING_OPERATORS,
            "experiment_id": STRING_OPERATORS,
            "start_time": DATE_TIME_OPERATORS,
            "end_time": DATE_TIME_OPERATORS,
            "created_at": DATE_TIME_OPERATORS,
            "last_updated_at": DATE_TIME_OPERATORS,
            "total_estimated_cost": NUMBER_OPERATORS,
            "llm_span_count": NUMBER_OPERATORS,
            "usage.total_tokens": NUMBER_OPERATORS,
            "usage.prompt_tokens": NUMBER_OPERATORS,
            "usage.completion_tokens": NUMBER_OPERATORS,
            "duration": NUMBER_OPERATORS,
            "input_json": DICTIONARY_OPERATORS,
            "output_json": DICTIONARY_OPERATORS,
            "metadata": DICTIONARY_OPERATORS,
            "feedback_scores": FEEDBACK_SCORES_OPERATORS,
            "span_feedback_scores": FEEDBACK_SCORES_OPERATORS,
            "tags": LIST_OPERATORS,
            "annotation_queue_ids": LIST_OPERATORS,
            "error_info": ["is_empty", "is_not_empty"],
            "default": STRING_OPERATORS,
        }

    @property
    def dictionary_fields(self) -> List[str]:
        return [
            "metadata",
            "input_json",
            "output_json",
            "feedback_scores",
            "span_feedback_scores",
        ]


class SpanOQLConfig(OQLConfig):
    """OQL configuration for span filtering.

    Based on backend's SpanField enum.
    See: apps/opik-backend/src/main/java/com/comet/opik/api/filter/SpanField.java
    """

    @property
    def columns(self) -> Dict[str, str]:
        return {
            "id": "string",
            "name": "string",
            "start_time": "date_time",
            "end_time": "date_time",
            "input": "string",
            "output": "string",
            "input_json": "dictionary",
            "output_json": "dictionary",
            "metadata": "dictionary",
            "model": "string",
            "provider": "string",
            "total_estimated_cost": "number",
            "tags": "list",
            "usage.total_tokens": "number",
            "usage.prompt_tokens": "number",
            "usage.completion_tokens": "number",
            "feedback_scores": "feedback_scores_number",
            "duration": "number",
            "error_info": "error_container",
            "type": "enum",
            "trace_id": "string",
        }

    @property
    def supported_operators(self) -> Dict[str, List[str]]:
        return {
            "id": STRING_OPERATORS,
            "name": STRING_OPERATORS,
            "input": STRING_OPERATORS,
            "output": STRING_OPERATORS,
            "model": STRING_OPERATORS,
            "provider": STRING_OPERATORS,
            "trace_id": STRING_OPERATORS,
            "type": ["=", "!="],
            "start_time": DATE_TIME_OPERATORS,
            "end_time": DATE_TIME_OPERATORS,
            "total_estimated_cost": NUMBER_OPERATORS,
            "usage.total_tokens": NUMBER_OPERATORS,
            "usage.prompt_tokens": NUMBER_OPERATORS,
            "usage.completion_tokens": NUMBER_OPERATORS,
            "duration": NUMBER_OPERATORS,
            "input_json": DICTIONARY_OPERATORS,
            "output_json": DICTIONARY_OPERATORS,
            "metadata": DICTIONARY_OPERATORS,
            "feedback_scores": FEEDBACK_SCORES_OPERATORS,
            "tags": LIST_OPERATORS,
            "error_info": ["is_empty", "is_not_empty"],
            "default": STRING_OPERATORS,
        }

    @property
    def dictionary_fields(self) -> List[str]:
        return ["metadata", "input_json", "output_json", "feedback_scores"]


class ThreadOQLConfig(OQLConfig):
    """OQL configuration for thread filtering.

    Based on backend's TraceThreadField enum.
    See: apps/opik-backend/src/main/java/com/comet/opik/api/filter/TraceThreadField.java
    """

    @property
    def columns(self) -> Dict[str, str]:
        return {
            "id": "string",
            "first_message": "string",
            "last_message": "string",
            "number_of_messages": "number",
            "duration": "number",
            "created_at": "date_time",
            "last_updated_at": "date_time",
            "start_time": "date_time",
            "end_time": "date_time",
            "feedback_scores": "feedback_scores_number",
            "status": "enum",
            "tags": "list",
            "annotation_queue_ids": "list",
        }

    @property
    def supported_operators(self) -> Dict[str, List[str]]:
        return {
            "id": STRING_OPERATORS,
            "first_message": STRING_OPERATORS,
            "last_message": STRING_OPERATORS,
            "number_of_messages": NUMBER_OPERATORS,
            "duration": NUMBER_OPERATORS,
            "created_at": DATE_TIME_OPERATORS,
            "last_updated_at": DATE_TIME_OPERATORS,
            "start_time": DATE_TIME_OPERATORS,
            "end_time": DATE_TIME_OPERATORS,
            "feedback_scores": FEEDBACK_SCORES_OPERATORS,
            "status": ["=", "!="],
            "tags": LIST_OPERATORS,
            "annotation_queue_ids": LIST_OPERATORS,
            "default": STRING_OPERATORS,
        }

    @property
    def dictionary_fields(self) -> List[str]:
        return ["feedback_scores"]


class DatasetItemOQLConfig(OQLConfig):
    """OQL configuration for dataset item filtering.

    Based on backend's DatasetItemField enum and FilterQueryBuilder.
    See: apps/opik-backend/src/main/java/com/comet/opik/api/filter/DatasetItemField.java
    """

    @property
    def columns(self) -> Dict[str, str]:
        # Maps to DatasetItemField enum values and their FieldType
        return {
            "id": "string",  # FieldType.STRING
            "data": "map",  # FieldType.MAP - supports nested key access
            "full_data": "string",  # FieldType.STRING - toString(data)
            "source": "string",  # FieldType.STRING
            "trace_id": "string",  # FieldType.STRING
            "span_id": "string",  # FieldType.STRING
            "tags": "list",  # FieldType.LIST
            "created_at": "date_time",  # FieldType.DATE_TIME
            "last_updated_at": "date_time",  # FieldType.DATE_TIME
            "created_by": "string",  # FieldType.STRING
            "last_updated_by": "string",  # FieldType.STRING
        }

    @property
    def supported_operators(self) -> Dict[str, List[str]]:
        return {
            "id": STRING_OPERATORS,
            "full_data": STRING_OPERATORS,
            "source": STRING_OPERATORS,
            "trace_id": STRING_OPERATORS,
            "span_id": STRING_OPERATORS,
            "created_by": STRING_OPERATORS,
            "last_updated_by": STRING_OPERATORS,
            "data": ["=", "!=", "contains", "not_contains", "starts_with", "ends_with"],
            "tags": LIST_OPERATORS,
            "created_at": DATE_TIME_OPERATORS,
            "last_updated_at": DATE_TIME_OPERATORS,
            "default": STRING_OPERATORS,
        }

    @property
    def dictionary_fields(self) -> List[str]:
        # Fields that support nested key access via dot notation (data.key_name)
        return ["data"]


class PromptVersionOQLConfig(OQLConfig):
    """OQL configuration for prompt version filtering."""

    @property
    def columns(self) -> Dict[str, str]:
        return {
            "id": "string",
            "commit": "string",
            "template": "string",
            "change_description": "string",
            "metadata": "dictionary",
            "type": "string",
            "tags": "list",
            "created_at": "date_time",
            "created_by": "string",
        }

    @property
    def supported_operators(self) -> Dict[str, List[str]]:
        return {
            "id": STRING_OPERATORS,
            "commit": STRING_OPERATORS,
            "template": STRING_OPERATORS,
            "change_description": STRING_OPERATORS,
            "metadata": DICTIONARY_OPERATORS,
            "type": ["=", "!="],
            "tags": LIST_OPERATORS,
            "created_at": DATE_TIME_OPERATORS,
            "created_by": STRING_OPERATORS,
            "default": STRING_OPERATORS,
        }

    @property
    def dictionary_fields(self) -> List[str]:
        return ["metadata"]


OPERATORS_WITHOUT_VALUES = {"is_empty", "is_not_empty"}

_DEFAULT_CONFIG = TraceOQLConfig()


class OpikQueryLanguage:
    """
    This method implements a parser that can be used to convert a filter string into a list of filters that the BE expects.

    For example, this class allows you to convert the query string: `input contains "hello"` into
    `[{'field': 'input', 'operator': 'contains', 'value': 'hello'}]` as expected by the BE.

    When converting a query string into another format, a common approach is:
    1. First convert the string into a series of tokens using a tokenizer
    2. Convert the list of tokens into an Abstract Syntax Tree (AST) using a parser
    3. Traverse the AST and convert it into the desired format using a formatter

    Due to the simple nature of the queries we currently support (no support for and/or operators, etc.), we have
    combined the tokenizer and formatter steps into a single parse method.

    The parse method works by iterating over the string character by character and extracting / validating the tokens.
    """

    def __init__(self, query_string: Optional[str], config: Optional[OQLConfig] = None):
        self.query_string = query_string or ""
        self._config = config or _DEFAULT_CONFIG

        self._cursor = 0
        self._filter_expressions = self._parse_expressions()
        self.parsed_filters = None
        if self._filter_expressions is not None:
            self.parsed_filters = json.dumps(self._filter_expressions)

    @classmethod
    def for_traces(cls, query_string: Optional[str]) -> "OpikQueryLanguage":
        """
        Creates a parser for filtering traces using OQL syntax. Returns an
        OpikQueryLanguage instance preconfigured with TraceOQLConfig that validates
        trace-specific fields. Empty or None query_string yields no filters;
        malformed queries raise ValueError during parsing.
        """
        return cls(query_string, TraceOQLConfig())

    @classmethod
    def for_spans(cls, query_string: Optional[str]) -> "OpikQueryLanguage":
        """
        Creates a parser for filtering spans using OQL syntax. Returns an
        OpikQueryLanguage instance preconfigured with SpanOQLConfig that validates
        span-specific fields. Empty or None query_string yields no filters;
        malformed queries raise ValueError during parsing.
        """
        return cls(query_string, SpanOQLConfig())

    @classmethod
    def for_threads(cls, query_string: Optional[str]) -> "OpikQueryLanguage":
        """
        Creates a parser for filtering trace threads using OQL syntax. Returns an
        OpikQueryLanguage instance preconfigured with ThreadOQLConfig that validates
        thread-specific fields. Empty or None query_string yields no filters;
        malformed queries raise ValueError during parsing.
        """
        return cls(query_string, ThreadOQLConfig())

    @classmethod
    def for_dataset_items(cls, query_string: Optional[str]) -> "OpikQueryLanguage":
        """
        Creates a parser for filtering dataset items using OQL syntax. Use this when working with
        dataset views or filtering items within a dataset. Returns an OpikQueryLanguage instance
        preconfigured with DatasetItemOQLConfig that validates dataset-specific fields like input,
        expected_output, and item metadata. Empty or None query_string yields no filters; malformed
        queries raise ValueError during parsing.
        """
        return cls(query_string, DatasetItemOQLConfig())

    @classmethod
    def for_prompt_versions(cls, query_string: Optional[str]) -> "OpikQueryLanguage":
        """
        Creates a parser for filtering prompt versions using OQL syntax. Use this when searching
        or filtering prompt version history. Returns an OpikQueryLanguage instance preconfigured
        with PromptVersionOQLConfig that validates prompt version fields like tags, template,
        commit, metadata, and created_at. Empty or None query_string yields no filters; malformed
        queries raise ValueError during parsing.
        """
        return cls(query_string, PromptVersionOQLConfig())

    def get_filter_expressions(self) -> Optional[List[Dict[str, Any]]]:
        return self._filter_expressions

    def _is_valid_field_char(self, char: str) -> bool:
        return char.isalnum() or char == "_"

    def _is_valid_connector_char(self, char: str) -> bool:
        return char.isalpha()

    def _skip_whitespace(self) -> None:
        while (
            self._cursor < len(self.query_string)
            and self.query_string[self._cursor].isspace()
        ):
            self._cursor += 1

    def _check_escaped_key(self) -> Tuple[bool, str]:
        if self.query_string[self._cursor] in ('"', "'"):
            is_quoted_key = True
            quote_type = self.query_string[self._cursor]
            self._cursor += 1
        else:
            is_quoted_key = False
            quote_type = ""

        return is_quoted_key, quote_type

    def _is_valid_escaped_key_char(self, quote_type: str, start: int) -> bool:
        if self.query_string[self._cursor] != quote_type:
            # Check this isn't the end of the string (means we missed the closing quote)
            if self._cursor + 2 >= len(self.query_string):
                raise ValueError(
                    "Missing closing quote for: " + self.query_string[start - 1 :]
                )

            return True

        # Check if it's an escaped quote (doubled quote)
        if (
            self._cursor + 1 < len(self.query_string)
            and self.query_string[self._cursor + 1] == quote_type
        ):
            # Skip the second quote
            self._cursor += 1
            return True

        return False

    def _parse_connector(self) -> str:
        start = self._cursor
        while self._cursor < len(self.query_string) and self._is_valid_connector_char(
            self.query_string[self._cursor]
        ):
            self._cursor += 1
        connector = self.query_string[start : self._cursor]
        return connector

    def _parse_field(self) -> Dict[str, Any]:
        # Skip whitespace
        self._skip_whitespace()

        columns = self._config.columns
        dictionary_fields = self._config.dictionary_fields

        # Parse the field name
        start = self._cursor
        while self._cursor < len(self.query_string) and self._is_valid_field_char(
            self.query_string[self._cursor]
        ):
            self._cursor += 1
        field = self.query_string[start : self._cursor]

        # Parse the key if it exists
        if (
            self._cursor < len(self.query_string)
            and self.query_string[self._cursor] == "."
        ):
            # Skip the "."
            self._cursor += 1

            # Check if the key is quoted
            is_quoted_key, quote_type = self._check_escaped_key()

            start = self._cursor
            while self._cursor < len(self.query_string) and (
                self._is_valid_field_char(self.query_string[self._cursor])
                or (
                    is_quoted_key and self._is_valid_escaped_key_char(quote_type, start)
                )
            ):
                self._cursor += 1

            key = self.query_string[start : self._cursor]

            # If escaped key, skip the closing quote
            if is_quoted_key:
                key = key.replace(
                    quote_type * 2, quote_type
                )  # Replace doubled quotes with single quotes
                self._cursor += 1

            # Special handling for usage.X fields (trace/span specific)
            # These are treated as flat fields, not dictionary access
            if field == "usage":
                composite_field = f"usage.{key}"
                if composite_field in columns:
                    return {
                        "field": composite_field,
                        "key": "",
                        "type": columns[composite_field],
                    }
                else:
                    raise ValueError(
                        f"When querying usage, {key} is not supported, only usage.total_tokens, usage.prompt_tokens and usage.completion_tokens are supported."
                    )

            # Keys are only supported for dictionary fields
            if field not in dictionary_fields:
                raise ValueError(
                    f"Field {field}.{key} is not supported, only the fields {list(columns.keys())} are supported."
                )
            elif field in columns:
                return {"field": field, "key": key, "type": columns[field]}
            else:
                # defaults to string
                return {"field": field, "key": key, "type": "string"}

        elif field in columns:
            return {"field": field, "key": "", "type": columns[field]}
        else:
            # defaults to string
            return {"field": field, "key": "", "type": "string"}

    def _parse_operator(self, parsed_field: str) -> Dict[str, Any]:
        # Skip whitespace
        self._skip_whitespace()

        supported_operators = self._config.supported_operators

        if parsed_field not in supported_operators:
            parsed_field = "default"

        # Parse the operator
        if self.query_string[self._cursor] == "=":
            operator = "="
            self._cursor += 1
            if operator not in supported_operators[parsed_field]:
                raise ValueError(
                    f"Operator {operator} is not supported for field {parsed_field}, only the operators {supported_operators[parsed_field]} are supported."
                )
            return {"operator": operator}

        elif self.query_string[self._cursor] in ["<", ">"]:
            if self.query_string[self._cursor + 1] == "=":
                operator = f"{self.query_string[self._cursor]}="
                self._cursor += 2
            else:
                operator = self.query_string[self._cursor]
                self._cursor += 1

            if operator not in supported_operators[parsed_field]:
                raise ValueError(
                    f"Operator {operator} is not supported for field {parsed_field}, only the operators {supported_operators[parsed_field]} are supported."
                )
            return {"operator": operator}
        else:
            start = self._cursor
            while (
                self._cursor < len(self.query_string)
                and not self.query_string[self._cursor].isspace()
            ):
                self._cursor += 1

            operator = self.query_string[start : self._cursor]
            if operator not in supported_operators[parsed_field]:
                raise ValueError(
                    f"Operator {operator} is not supported for field {parsed_field}, only the operators {supported_operators[parsed_field]} are supported."
                )
            return {"operator": operator}

    def _get_number(self) -> str:
        start = self._cursor
        while (
            self._cursor < len(self.query_string)
            and self.query_string[self._cursor].isdigit()
        ):
            self._cursor += 1
        return self.query_string[start : self._cursor]

    def _parse_value(self) -> Dict[str, Any]:
        self._skip_whitespace()

        start = self._cursor
        if self.query_string[self._cursor] == '"':
            self._cursor += 1
            start = self._cursor

            # TODO: replace with new quote parser used in field parser
            while (
                self._cursor < len(self.query_string)
                and self.query_string[self._cursor] != '"'
            ):
                self._cursor += 1

            value = self.query_string[start : self._cursor]

            # Add 1 to skip the closing quote and return the value
            self._cursor += 1
            return {"value": value}
        elif (
            self.query_string[self._cursor].isdigit()
            or self.query_string[self._cursor] == "-"
        ):
            value = self._get_number()
            if (
                self._cursor < len(self.query_string)
                and self.query_string[self._cursor] == "."
            ):
                self._cursor += 1
                value += "." + self._get_number()

            return {"value": value}
        else:
            raise ValueError(
                f'Invalid value {self.query_string[start : self._cursor]}, expected an string in double quotes("value") or a number'
            )

    def _parse_expressions(self) -> Optional[List[Dict[str, Any]]]:
        if len(self.query_string) == 0:
            return None

        expressions = []

        while True:
            # Parse fields
            parsed_field = self._parse_field()

            # Parse operators
            parsed_operator = self._parse_operator(parsed_field["field"])

            operator_name = parsed_operator.get("operator", "")
            if operator_name in OPERATORS_WITHOUT_VALUES:
                # For operators without values, use empty string as value
                parsed_value = {"value": ""}
            else:
                parsed_value = self._parse_value()

            expressions.append({**parsed_field, **parsed_operator, **parsed_value})

            self._skip_whitespace()

            if self._cursor < len(self.query_string):
                position = self._cursor
                connector = self._parse_connector()

                if connector.lower() == "and":
                    continue
                elif connector.lower() == "or":
                    raise ValueError(
                        "Invalid filter string, OR is not currently supported"
                    )
                else:
                    raise ValueError(
                        f"Invalid filter string, trailing characters {self.query_string[position:]}"
                    )
            else:
                break

        return expressions



"""Save structured JSON report files for evaluation suite results."""

from __future__ import annotations

import json
import logging
import os
from typing import Optional, TYPE_CHECKING

if TYPE_CHECKING:
    from .. import evaluation_suite_result as _result_mod

LOGGER = logging.getLogger(__name__)

DEFAULT_REPORT_DIR = "opik_evaluation_suite_reports"


def save_report(
    suite_result: _result_mod.EvaluationSuiteResult,
    output_path: Optional[str] = None,
) -> str:
    """Save an evaluation suite result as a structured JSON report file.

    Args:
        suite_result: The evaluation suite result to serialize.
        output_path: Optional file path. If not provided, a default path
            is generated under the ``opik_evaluation_suite_reports/`` directory.

    Returns:
        The absolute path to the written report file.
    """
    report_dict = suite_result.to_report_dict()

    if output_path is None:
        output_path = build_default_report_path(
            suite_result.experiment_name or suite_result.experiment_id
        )

    output_path = os.path.abspath(output_path)
    parent_dir = os.path.dirname(output_path)
    if parent_dir:
        os.makedirs(parent_dir, exist_ok=True)

    with open(output_path, "w", encoding="utf-8") as f:
        json.dump(report_dict, f, indent=2, default=str, ensure_ascii=False)

    LOGGER.debug("Evaluation suite report saved to %s", output_path)
    return output_path


def build_default_report_path(experiment_name: str) -> str:
    """Build a default report file path from an experiment name."""
    safe_name = _sanitize_filename(experiment_name)
    return os.path.abspath(os.path.join(DEFAULT_REPORT_DIR, f"{safe_name}.json"))


def _sanitize_filename(name: str) -> str:
    """Replace characters that are unsafe in file names."""
    return "".join(c if c.isalnum() or c in "-_." else "_" for c in name)



import dataclasses
import logging
from typing import Any, Callable, Dict, Optional

from google.adk import models as adk_models
from google.genai import types as genai_types

from .. import helpers as adk_helpers

import opik
from opik import llm_usage
from opik.llm_usage import opik_usage

LOGGER = logging.getLogger(__name__)


class LlmResponseCreateWrapper:
    def __init__(
        self,
        wrapped_response_create: Callable[
            [genai_types.GenerateContentResponse], adk_models.LlmResponse
        ],
    ) -> None:
        self.wrapped_response_create = wrapped_response_create

    def __call__(
        self, generate_content_response: genai_types.GenerateContentResponse
    ) -> adk_models.LlmResponse:
        LOGGER.debug("LlmResponseCreateWrapper called")
        return _wrap_llm_response_create(
            generate_content_response, self.wrapped_response_create
        )


@dataclasses.dataclass
class LLMUsageData:
    opik_usage: opik_usage.OpikUsage
    model: Optional[str]
    provider: Optional[str]


def pop_llm_usage_data(
    result_dict: Dict[str, Any], provider: Optional[str]
) -> Optional[LLMUsageData]:
    """Extracts Opik usage metadata from ADK output and removes it from the result dict."""
    opik_usage_metadata = None
    model = None

    if (custom_metadata := result_dict.get("custom_metadata", None)) is not None:
        if (opik_usage_metadata := custom_metadata.pop("opik_usage", None)) is not None:
            model = custom_metadata.pop("model_version", None)
            if model is not None:
                model = model.split("/")[-1]

    # in streaming mode ADK returns the usage metadata in the result dict as the last call
    # to the after_model_callback bypassing our patching (no opik_usage)
    if opik_usage_metadata is None:
        if "usage_metadata" in result_dict:
            opik_usage_metadata = result_dict["usage_metadata"]
        else:
            return None

    if provider in [opik.LLMProvider.GOOGLE_AI, opik.LLMProvider.GOOGLE_VERTEXAI]:
        usage = llm_usage.try_build_opik_usage_or_log_error(
            provider=opik.LLMProvider(provider),
            usage=opik_usage_metadata,
            logger=LOGGER,
            error_message="Failed to log token usage from ADK Gemini call",
        )
    else:
        usage = llm_usage.build_opik_usage_from_unknown_provider(
            usage=opik_usage_metadata,
        )

    if usage is None:
        return None

    return LLMUsageData(opik_usage=usage, model=model, provider=provider)


def _wrap_llm_response_create(
    generate_content_response: genai_types.GenerateContentResponse,
    wrapped_response_create: Callable[
        [genai_types.GenerateContentResponse], adk_models.LlmResponse
    ],
) -> adk_models.LlmResponse:
    usage_metadata = generate_content_response.usage_metadata
    LOGGER.debug("_wrap_llm_response_create: usage_metadata=%s", usage_metadata)

    response = wrapped_response_create(generate_content_response)
    if usage_metadata is None:
        return response

    if response.custom_metadata is None:
        response.custom_metadata = {}

    response.custom_metadata["opik_usage"] = usage_metadata
    response.custom_metadata["provider"] = adk_helpers.get_adk_provider()
    response.custom_metadata["model_version"] = generate_content_response.model_version
    LOGGER.debug(
        "_wrap_llm_response_create finished: response.custom_metadata=%s",
        response.custom_metadata,
    )

    return response



import logging
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple,
    Union,
)
from typing_extensions import override

from google.genai import types as genai_types

import opik.dict_utils as dict_utils
import opik.llm_usage as llm_usage
from opik.api_objects import span
from opik.decorator import arguments_helpers, base_track_decorator
from opik.types import LLMProvider

from . import stream_wrappers

LOGGER = logging.getLogger(__name__)


KWARGS_KEYS_TO_LOG_AS_INPUTS = ["contents", "config"]
RESPONSE_KEYS_TO_LOG_AS_OUTPUT = ["candidates"]


class GenerateContentTrackDecorator(base_track_decorator.BaseTrackDecorator):
    """
    An implementation of BaseTrackDecorator designed specifically for tracking
    calls to genai.Client's
    * models.generate_content
    * models.generate_content_stream
    * aio.models.generate_content
    * aio.models.generate_content_stream
    """

    def __init__(
        self,
        provider: str,
        cost_callback: Optional[
            Callable[[genai_types.GenerateContentResponse], Optional[float]]
        ] = None,
    ) -> None:
        super().__init__()
        self.provider = provider
        self._cost_callback = cost_callback

    @override
    def _start_span_inputs_preprocessor(
        self,
        func: Callable,
        track_options: arguments_helpers.TrackOptions,
        args: Tuple,
        kwargs: Dict[str, Any],
    ) -> arguments_helpers.StartSpanParameters:
        assert kwargs is not None, (
            "Expected kwargs to be not None in client.models.generate_content(**kwargs), client.aio.models.generate_content(**kwargs)"
        )

        model = kwargs.get("model")

        name = track_options.name if track_options.name is not None else func.__name__
        name = f"{name}: {model}"  # Add model to the name for better viewing UX

        metadata = track_options.metadata if track_options.metadata is not None else {}

        input, new_metadata = dict_utils.split_dict_by_keys(
            kwargs, keys=KWARGS_KEYS_TO_LOG_AS_INPUTS
        )
        metadata = dict_utils.deepmerge(metadata, new_metadata)
        metadata.update({"created_from": "genai"})

        tags = ["genai"]

        result = arguments_helpers.StartSpanParameters(
            name=name,
            input=input,
            type=track_options.type,
            tags=tags,
            metadata=metadata,
            project_name=track_options.project_name,
            model=model,
            provider=self.provider,
        )

        return result

    @override
    def _end_span_inputs_preprocessor(
        self,
        output: Any,
        capture_output: bool,
        current_span_data: span.SpanData,
    ) -> arguments_helpers.EndSpanParameters:
        assert isinstance(
            output,
            genai_types.GenerateContentResponse,
        ), f"{output}"

        result_dict = output.model_dump(mode="json")
        output, metadata = dict_utils.split_dict_by_keys(
            result_dict, RESPONSE_KEYS_TO_LOG_AS_OUTPUT
        )

        if result_dict.get("model_version") is not None:
            # Gemini **may** add "models/" prefix to some model versions
            model = result_dict["model_version"].split("/")[-1]
        else:
            model = None

        usage = llm_usage.try_build_opik_usage_or_log_error(
            provider=LLMProvider(self.provider),
            usage=result_dict["usage_metadata"],
            logger=LOGGER,
            error_message="Failed to log token usage from genai generate_response call",
        )

        total_cost = None
        if self._cost_callback is not None:
            total_cost = self._cost_callback(output)

        span_name_without_model = current_span_data.name.split(":")[0]  # type: ignore
        result = arguments_helpers.EndSpanParameters(
            name=f"{span_name_without_model}: {model}",
            output=output,
            usage=usage,
            metadata=metadata,
            model=model,
            provider=self.provider,
            total_cost=total_cost,
        )

        return result

    @override
    def _streams_handler(
        self,
        output: Any,
        capture_output: bool,
        generations_aggregator: Optional[Callable[[List[Any]], Any]],
    ) -> Union[
        None,
        Iterator[genai_types.GenerateContentResponse],
        AsyncIterator[genai_types.GenerateContentResponse],
    ]:
        if isinstance(output, Iterator):
            span_to_end, trace_to_end = base_track_decorator.pop_end_candidates()
            assert generations_aggregator is not None
            return stream_wrappers.wrap_sync_iterator(
                stream=output,
                span_to_end=span_to_end,
                trace_to_end=trace_to_end,
                generations_aggregator=generations_aggregator,
                finally_callback=self._after_call,
            )

        if isinstance(output, AsyncIterator):
            span_to_end, trace_to_end = base_track_decorator.pop_end_candidates()
            assert generations_aggregator is not None
            return stream_wrappers.wrap_async_iterator(
                stream=output,
                span_to_end=span_to_end,
                trace_to_end=trace_to_end,
                generations_aggregator=generations_aggregator,
                finally_callback=self._after_call,
            )

        NO_STREAM_DETECTED = None
        return NO_STREAM_DETECTED



from typing import Callable, Optional

import google.genai as genai
from google.genai import types as genai_types

from . import (
    generate_content_decorator,
    generations_aggregators,
    stream_wrappers,
    encoder_extension,
)


def _get_provider(client: genai.Client) -> str:
    """Get the provider name from the GenAI client."""
    return "google_vertexai" if client.vertexai else "google_ai"


def track_genai(
    client: genai.Client,
    project_name: Optional[str] = None,
    upload_videos: bool = True,
    cost_callback: Optional[
        Callable[[genai_types.GenerateContentResponse], Optional[float]]
    ] = None,
) -> genai.Client:
    """
    Adds Opik tracking to an genai.Client.

    Tracks calls to:
    * client.models.generate_content
    * client.models.generate_content_stream
    * client.aio.models.generate_content
    * client.aio.models.generate_content_stream
    * client.models.generate_videos (Veo video generation)
    * client.aio.models.generate_videos (async Veo video generation)
    * client.operations.get (for polling video generation status)
    * video.save (when saving generated videos)

    Can be used within other Opik-tracked functions.

    Args:
        client: An instance of genai.Client.
        project_name: The name of the project to log data.
        upload_videos: Whether to upload generated videos as attachments
            when video.save is called. Defaults to True.
        cost_callback: An optional callable that receives the GenerateContentResponse
            and returns the cost in USD as a float, or None to use automatic cost
            calculation. Use this for models with complex pricing (e.g., different
            rates for thinking tokens) where automatic cost tracking is insufficient.

    Returns:
        The modified genai.Client with Opik tracking enabled.
    """
    if hasattr(client, "opik_tracked"):
        return client
    encoder_extension.register()

    client.opik_tracked = True

    provider = _get_provider(client)

    _patch_generate_content(client, provider, project_name, cost_callback)
    _patch_generate_videos(client, provider, project_name, upload_videos)

    return client


def _patch_generate_content(
    client: genai.Client,
    provider: str,
    project_name: Optional[str],
    cost_callback: Optional[
        Callable[[genai_types.GenerateContentResponse], Optional[float]]
    ] = None,
) -> None:
    """Patch generate_content methods with Opik tracking."""
    decorator_factory = generate_content_decorator.GenerateContentTrackDecorator(
        provider=provider,
        cost_callback=cost_callback,
    )

    client.models.generate_content = decorator_factory.track(
        name="generate_content",
        type="llm",
        project_name=project_name,
    )(client.models.generate_content)

    # This conversion is needed so that @track decorator stopped treating generate_content_stream
    # as generator function, as they are tracked slightly differently from what we need here.
    # The user still gets their Iterator object in return.
    client.models.generate_content_stream = (
        stream_wrappers.generator_function_to_normal_function(
            client.models.generate_content_stream
        )
    )
    client.models.generate_content_stream = decorator_factory.track(
        name="generate_content_stream",
        type="llm",
        project_name=project_name,
        generations_aggregator=generations_aggregators.aggregate_response_content_items,
    )(client.models.generate_content_stream)

    client.aio.models.generate_content = decorator_factory.track(
        name="async_generate_content",
        type="llm",
        project_name=project_name,
    )(client.aio.models.generate_content)

    # No need to perform a similar conversion as for the synchronous method, because async version of generate_content_stream
    # is already a wrapper around generator function that works similarly to our helper function
    client.aio.models.generate_content_stream = decorator_factory.track(
        name="async_generate_content_stream",
        type="llm",
        project_name=project_name,
        generations_aggregator=generations_aggregators.aggregate_response_content_items,
    )(client.aio.models.generate_content_stream)


GENAI_VIDEOS_TAGS = ["genai"]
GENAI_VIDEOS_METADATA = {"created_from": "genai", "type": "genai_videos"}


def _patch_generate_videos(
    client: genai.Client,
    provider: str,
    project_name: Optional[str],
    upload_videos: bool,
) -> None:
    """Patch generate_videos and operations.get methods with Opik tracking."""
    from . import videos

    if not hasattr(client.models, "generate_videos"):
        return

    video_decorator_factory = videos.GenerateVideosTrackDecorator(provider=provider)

    # Patch sync generate_videos
    client.models.generate_videos = video_decorator_factory.track(
        name="models.generate_videos",
        type="llm",
        project_name=project_name,
        tags=GENAI_VIDEOS_TAGS,
        metadata=GENAI_VIDEOS_METADATA,
    )(client.models.generate_videos)

    # Patch async generate_videos
    if hasattr(client.aio.models, "generate_videos"):
        client.aio.models.generate_videos = video_decorator_factory.track(
            name="models.generate_videos",
            type="llm",
            project_name=project_name,
            tags=GENAI_VIDEOS_TAGS,
            metadata=GENAI_VIDEOS_METADATA,
        )(client.aio.models.generate_videos)

    # Patch operations.get to track polling and patch Video.save on completed videos
    _patch_operations_get(client, project_name, upload_videos)


def _patch_operations_get(
    client: genai.Client, project_name: Optional[str], upload_videos: bool
) -> None:
    """Patch operations.get to track polling and patch Video.save on completed videos."""
    from . import videos

    if not hasattr(client, "operations") or not hasattr(client.operations, "get"):
        return

    operations_decorator = videos.OperationsGetTrackDecorator(
        project_name=project_name, upload_videos=upload_videos
    )
    client.operations.get = operations_decorator.track(
        name="operations.get",
        type="general",
        project_name=project_name,
        tags=GENAI_VIDEOS_TAGS,
        metadata=GENAI_VIDEOS_METADATA,
    )(client.operations.get)



from typing import Any, Callable


def patch(instance: Any, decorator: Callable[[Callable], Callable]) -> None:
    attr_name: str
    for attr_name in dir(instance):
        attr_value = getattr(instance, attr_name)
        if callable(attr_value) and not attr_name.startswith("_"):
            decorated_method = decorator(attr_value)
            setattr(instance, attr_name, decorated_method)
